/*
**   IRC - Internet Relay Chat, source/misc.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
#include "misc.h"

RCSID("$Id: misc.c,v 1.33 1999/02/23 07:57:22 cbehrens Exp $");

typedef struct _command			COMMAND;

struct _command
{
	char			*command;
	commandfunc_t	function;
};

typedef struct _commandentry	COMMANDENTRY;

struct _commandentry
{
	COMMAND			*mptr;
	COMMANDENTRY	*next[27];
};

static COMMANDENTRY *_commandtree[27];
static COMMAND	_Commands[] =
{
	{	"SET",		m_set		},
	{	"GET",		m_get		},
	{	"PRIVMSG",	m_privmsg	},
	{	"WHO",		m_who		},
	{	"WHOIS",	m_whois		},
	{	"WHOWAS",	m_whowas	},
	{	"USER",		m_user		},
	{	"NICK",		m_nick		},
	{	"SERVER",	m_server	},
	{	"LIST",		m_list		},
	{	"TOPIC",	m_topic		},
	{	"INVITE",	m_invite	},
	{	"VERSION",	m_version	},
	{	"QUIT",		m_quit		},
	{	"SQUIT",	m_squit		},
	{	"KILL",		m_kill		},
	{	"INFO",		m_info		},
	{	"LINKS",	m_links		},
	{	"STATS",	m_stats		},
	{	"HELP",		m_help		},
	{	"ERROR",	m_error		},
	{	"AWAY",		m_away		},
	{	"CONNECT",	m_connect	},
	{	"PING",		m_ping		},
	{	"PONG",		m_pong		},
	{	"OPER",		m_oper		},
	{	"PASS",		m_pass		},
	{	"WALLOPS",	m_wallops	},
	{	"TIME",		m_time		},
	{	"NAMES",	m_names		},
	{	"ADMIN",	m_admin		},
	{	"TRACE",	m_trace		},
	{	"NOTICE",	m_notice	},
	{	"JOIN",		m_join		},
	{	"PART",		m_part		},
	{	"LUSERS",	m_lusers	},
	{	"MOTD",		m_motd		},
	{	"MODE",		m_mode		},
	{	"KICK",		m_kick		},
	{	"USERHOST",	m_userhost	},
	{	"ISON",		m_ison		},
	{	"NOTE",		m_note		},
#ifdef USE_SERVICES
	{	"SERVICE",	m_service	},
	{	"SQUERY",	m_squery	},
	{	"SERVLIST",	m_servlist	},
	{	"SERVSET",	m_servset	},
#endif
	{	"REHASH",	m_rehash	},
	{	"RESTART",	m_restart	},
	{	"CLOSE",	m_close		},
	{	"SVINFO",	m_svinfo	},
	{	"SJOIN",	m_sjoin		},
	{	"DIE",		m_die		},
	{	"DNS",		m_dns		},
	{	"GLINE",	m_gline		},
	{	"EXT",		m_ext		},
	{	"OPERWALL",	m_operwall	},
	{	"KLINE",	m_kline		},
	{	"QUEUE",	m_queue		},
	{	"FLAGS",	m_flags		},
	{	"IPHASH",	m_iphash	},
	{	"CAPAB",	m_capab		},
#ifdef MALLOC_DEBUG
	{	"MALLOC",	m_malloc	},
#endif
	{	"REJECTHOLD",	m_rejecthold		},
	{	NULL,		NULL		}
};

static void _command_tree_add(char *name, COMMAND *cptr);
static COMMAND *_command_tree_find(char *name);
static int _client_del_remote(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason, int squits, char *clireason);
static int _client_del_local(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason);

static void _command_tree_add(char *name, COMMAND *cptr)
{
	char			*ptr = name;
	COMMANDENTRY	*lastentry = NULL;
	COMMANDENTRY	*entry;
	unsigned int	num;

	if (!*ptr)
		return;
	do
	{
		if (!*ptr)
			num = 0;
		else if (islower(*ptr))
			num = toupper(*ptr) - 'A' + 1;
		else
			num = *ptr - 'A' + 1;
		if (num >= 27)
			return;
		if (!lastentry)
			entry = _commandtree[num];
		else
			entry = lastentry->next[num];
		if (!entry)
   		{
			entry = (COMMANDENTRY *)malloc(sizeof(COMMANDENTRY));
			BZERO(entry, sizeof(COMMANDENTRY));
			if (!lastentry)
				_commandtree[num] = entry;
			else  
				lastentry->next[num] = entry;
		}
		lastentry = entry;
	} while(*ptr++); 
	if (entry->mptr)
		return;
	entry->mptr = cptr;
}

static COMMAND *_command_tree_find(char *name)
{
	char			*ptr = name;
	COMMANDENTRY	*lastentry = NULL;
	COMMANDENTRY	*entry;
	unsigned int	num;

	if (!*ptr)
		return NULL;
	do
	{
		if (!*ptr)
			num = 0;
		else if (islower(*ptr))
			num = toupper(*ptr) - 'A' + 1;
        else
            num = *ptr - 'A' + 1;
		if (num >= 27)
			return NULL;
		if (!lastentry)
			entry = _commandtree[num];
		else
			entry = lastentry->next[num];
		lastentry = entry;
		if (!entry)
			return NULL;
	} while(*ptr++);
	return entry->mptr;
}

void command_tree_init(void)
{
	COMMAND	*cptr = _Commands;
	int		i;

	for(i=0;i<27;i++)
		_commandtree[i] = NULL;
	for(;cptr->command;cptr++)
		_command_tree_add(cptr->command, cptr);
}

int ircd_open_fd(int fd, int flags, IRCD_CLIENT *ptr, FDLIST *fdlistptr)
{
	FDLISTINFO	*fdlistinfo;

	if (fd < 0)
		return -EINVAL;
	SET_FDS(fd, ptr);
	fdlistinfo = &(ptr->fdlistinfo);
	fdlistinfo->fd = fd;
	fdlistinfo->fdlist = fdlistptr ? fdlistptr : &Fdlist_main;
	if (flags)
	{
		fdlistinfo->fdlist_pos = fdlist_add_fd(fdlistinfo->fdlist, fd, flags);
		return fdlistinfo->fdlist_pos;
	}
	fdlistinfo->fdlist_pos = -1;
	return 0;
}

int ircd_close_fd(FDLISTINFO *fdlistinfo)
{
	int		ret;
	int		mfd;

	if (!fdlistinfo || fdlistinfo->fd < 0)
		return -EINVAL;
	close(fdlistinfo->fd);
	SET_FDS(fdlistinfo->fd, NULL);
	if (fdlistinfo->fdlist && fdlistinfo->fdlist_pos >= 0)
	{
		ret = fdlist_del_fd(fdlistinfo->fdlist, fdlistinfo->fdlist_pos, &mfd);
		if ((ret > 0) && (mfd >= 0)) /* update moved fds's position in fdlist */
		{
			if (GET_FDS(mfd))
				GET_FDS(mfd)->fdlistinfo.fdlist_pos = fdlistinfo->fdlist_pos;
		}
	}
	return 0;
}

int client_add_rejecthold(int fd, struct sockaddr_in *rem_sin)
{
	IRCD_CLIENT	*cptr;
	int			ret;

	if ((ret = mysocket_set_nonblocking(fd)) < 0)
		return ret;
	cptr = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
	if (!cptr)
		return -ENOMEM;
    cptr->rem_ipaddr = ntohl(rem_sin->sin_addr.s_addr);
    cptr->rem_port = ntohs(rem_sin->sin_port);
	cptr->lasttime = cptr->firsttime = Now;
    ret = queue_init(&(cptr->queue_in), QUEUE_SIZE, IQUEUE_FLAGS_USEMMAP);
    if (ret < 0)
    {
        MyFree(cptr);
        return ret;
    }
    ret = queue_init(&(cptr->queue_out), QUEUE_SIZE, IQUEUE_FLAGS_USEMMAP);
    if (ret < 0)
    {
        queue_delete(&(cptr->queue_in));
        MyFree(cptr);
        return ret;
    }
    ret = ircd_open_fd(fd, FDLIST_FLAGS_READ, cptr, NULL);
    if (ret < 0)
    {
        queue_delete(&(cptr->queue_in));
        queue_delete(&(cptr->queue_out));
        MyFree(cptr);
        return ret;
    }
	cptr->flags |= CLIENT_FLAGS_LOCAL;
	add_client_to_lolist(&Rejectholds, cptr);
	cptr->type = CLIENT_TYPE_REJECTHOLD;
	return 0;
}

int client_add_local(IRCD_CLIENT *listener, int fd, struct sockaddr_in *rem_sin)
{
    int                 ret;
    struct sockaddr_in  loc_sin;
    int                 sin_size = sizeof(loc_sin);
    IRCD_CLIENT         *cptr;
	IRCD_CLIENT			*acptr;
	int					afd;
	MYRES_HOSTENT		*heptr;
	struct in_addr		inaddr;

	ret = getsockname(fd, (struct sockaddr *)(&loc_sin), &sin_size);
	if (ret < 0)
		return -errno;
	if ((ret = mysocket_set_nonblocking(fd)) < 0)
		return ret;
	cptr = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
	if (!cptr)
		return -ENOMEM;
	cptr->maxsendq = 8192;
	cptr->from = cptr;
    cptr->rem_ipaddr = ntohl(rem_sin->sin_addr.s_addr);
    cptr->rem_port = ntohs(rem_sin->sin_port);
    cptr->loc_ipaddr = ntohl(loc_sin.sin_addr.s_addr);
    cptr->loc_port = ntohs(loc_sin.sin_port);
  	strcpy(cptr->ipstring, inet_ntoa(rem_sin->sin_addr));
	cptr->listener = listener;
	cptr->lasttime = cptr->firsttime = Now;
	cptr->idletime = Now;

	strcpy(cptr->username, "unknown");
	strncpy(cptr->hostname, inet_ntoa(rem_sin->sin_addr),
				sizeof(cptr->hostname)-1);

    ret = queue_init(&(cptr->queue_in), QUEUE_SIZE, IQUEUE_FLAGS_USEMMAP);
    if (ret < 0)
    {
        MyFree(cptr);
        return ret;
    }
    ret = queue_init(&(cptr->queue_out), QUEUE_SIZE, IQUEUE_FLAGS_USEMMAP);
    if (ret < 0)
    {
        queue_delete(&(cptr->queue_in));
        MyFree(cptr);
        return ret;
    }
    ret = ircd_open_fd(fd, FDLIST_FLAGS_READ, cptr, NULL);
    if (ret < 0)
    {
        queue_delete(&(cptr->queue_in));
        queue_delete(&(cptr->queue_out));
        MyFree(cptr);
        return ret;
    }
	add_client_to_lolist(&Locals, cptr);
	cptr->flags |= CLIENT_FLAGS_LOCAL;
	add_client_to_llist(&(listener->listener), cptr);
	listener->num_on++;
	myres_status_init(&(cptr->res_status), (void *)cptr, MYRES_TYPE_CLIENT);

	inaddr.s_addr = htonl(cptr->rem_ipaddr);
	heptr = myres_gethostbyaddr(&inaddr, &(cptr->res_status));
	if (heptr)
	{
		myres_hostent_move(&(cptr->he_reverse), &(cptr->res_status.he));
		MyFree(heptr);
		cptr->res_status.type = 0;
		heptr = myres_gethostbyname(cptr->he_reverse.h.h_name, &(cptr->res_status));
		if (heptr)
		{
			myres_hostent_move(&(cptr->he_forward), &(cptr->res_status.he));
			MyFree(heptr);
			myres_status_deinit(&(cptr->res_status));
		}
		else
			cptr->flags |= CLIENT_FLAGS_DOINGDNS;
	}
	else
		cptr->flags |= CLIENT_FLAGS_DOINGDNS;

	afd = mysocket_connectip(cptr->rem_ipaddr, 113,
		cptr->loc_ipaddr, MYSOCKET_FLAGS_REUSEADDR|MYSOCKET_FLAGS_NONBLOCKING);
	if (afd >= 0)
	{
		if (afd >= MAXCONNECTIONS)
		{
			close(afd);
			return 0;
		}
		acptr = cptr->auth_ptr = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
		acptr->lasttime = acptr->firsttime = Now;
		acptr->type = CLIENT_TYPE_AUTH;
		acptr->flags = CLIENT_FLAGS_LOCAL;
		acptr->auth_ptr = cptr;
		cptr->flags |= CLIENT_FLAGS_DOINGAUTH;
		acptr->flags |= CLIENT_FLAGS_DOINGCONNECT;
	    ret = queue_init(&(acptr->queue_in), QUEUE_SIZE, IQUEUE_FLAGS_USEMMAP);
	    if (ret < 0)
	    {
			close(afd);
	        MyFree(acptr);
			cptr->auth_ptr = NULL;
			cptr->flags &= ~CLIENT_FLAGS_DOINGAUTH;
	        return 0;
	    }
		if (ircd_open_fd(afd, FDLIST_FLAGS_READ|FDLIST_FLAGS_WRITE, acptr, NULL) < 0)
		{
			queue_delete(&(acptr->queue_in));
			close(afd);
			MyFree(acptr);
			cptr->auth_ptr = NULL;
			cptr->flags &= ~CLIENT_FLAGS_DOINGAUTH;
			return 0;
		}
		add_client_to_lolist(&Locals, acptr);
	}
    return 0;
}

int client_del(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason)
{
	if (IsAuth(cptr))
	{
		_client_del_local(from, cptr, reason);
		return -1;
	}
	else if (IsLocal(cptr))
	{
		char	cnamebuf1[CLIENTNAMESIZE];

		if (!(cptr->flags & CLIENT_FLAGS_NOSENDQUIT))
		{
			if (from != NULL && from != cptr)
				send_to_client(cptr, "ERROR :Closing Link: %s %s (%s)",
	                   ircd_clientname(cptr,0,cnamebuf1),
	                   from->name, reason);
			else
				send_to_client(cptr, "ERROR :Closing Link: %s (%s)",
	                   ircd_clientname(cptr,0,cnamebuf1), reason);
		}
		_client_del_local(from, cptr, reason);
		return -1;
	}
	if (IsServer(cptr))
	{
		char	comment[HOSTLEN*2+2];

		strcpy(comment, cptr->servptr->name);
		strcat(comment, " ");
		strcat(comment, cptr->name);
		if (from->from != cptr->from)
			send_to_client(cptr, ":%s SQUIT %s :%s", from->name, cptr->name, reason);
#ifndef REMOVE_CLIENTS_IMMEDIATELY
		else
#endif
			_client_del_remote(from, cptr, reason, 1, comment);
	}
	else
		_client_del_remote(from, cptr, reason, 1, NULL);
	return 0;
}

static int _client_del_remote(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason, int squits, char *clireason)
{
	/* remotes are either servers or clients*/
	IRCD_CLIENT	**servbucket;
	IRCD_CLIENT	**userbucket;
	IRCD_CLIENT	*tempuser;
	IRCD_CLIENT	*nextuser;
	IRCD_CLIENT	*acptr;
	char		cnamebuf1[CLIENTNAMESIZE];

	assert(!IsLocal(cptr) || IsServer(cptr));
	if (!clireason)
		clireason = reason;
	servbucket = &(cptr->servers);
	userbucket = &(cptr->users);
	if (cptr->oslist)
	{
		del_client_from_oslist(cptr->oslist, cptr);
		cptr->oslist = NULL;
	}
	if (cptr->ipaddr_ptr)
	{
		iphash_handle_exit(cptr);
	}
	if (IsServer(cptr))
	{
		for(tempuser=*servbucket;tempuser;tempuser=nextuser)
		{
			nextuser = tempuser->u_next;
			_client_del_remote(from, tempuser, reason, 0, clireason);
		}
		for(tempuser = *userbucket;tempuser;tempuser=nextuser)
		{
			nextuser = tempuser->u_next;
			assert(tempuser->servptr == cptr);
			_client_del_remote(from, tempuser, reason, 0, clireason);
		}
		for(acptr=meptr->servers;acptr;acptr=acptr->u_next)
		{
			if (acptr == cptr)
				continue;
			if (!match(conf_myname_for_link(meptr->name,
						acptr->hostmask, cnamebuf1), cptr->name))
				continue;
			if (cptr->from == acptr)
				continue; /* used to have :prefix squit .. here */
			else
			{
				if (squits ||
						!IsCapable(acptr, IRCD_CAPAB_VALUE_NOQUITSTORM))
				send_to_client(acptr, "SQUIT %s :%s",
					cptr->name, reason);
			}
		}
		*servbucket = NULL;
		*userbucket = NULL;
		hash_del_client(cptr);
		Globalinfo.g_servers--;
		if (cptr->servptr) /* has to be true ? */
		{
			del_client_from_ulist(&(cptr->servptr->servers), cptr);
			cptr->servptr = NULL;
		}
		MyFree(cptr);
	}
	else
	{
		whowas_add_history(cptr, 0);
		whowas_off_history(cptr);
		if (!(cptr->flags & CLIENT_FLAGS_NOSENDQUIT))
		{
			IRCD_CLIENT	*ssptr;
			for(ssptr=meptr->servers;ssptr;ssptr=ssptr->u_next)
			{
				if (cptr->from == ssptr)
					continue;
				if (!IsCapable(ssptr, IRCD_CAPAB_VALUE_NOQUITSTORM) ||
					(ssptr->hostmask &&
					!match(conf_myname_for_link(meptr->name,
							ssptr->hostmask, cnamebuf1), cptr->servername)))
				send_to_client(ssptr, ":%s QUIT :%s",
					cptr->name, clireason);
			}
		}
		send_to_common_channels(cptr, "QUIT :%s", clireason);
		if (cptr->umodes & CLIENT_UMODES_OPER)
			Globalinfo.g_opers--;
		if (cptr->umodes & CLIENT_UMODES_IMODE)
			Globalinfo.g_invis--;
		Globalinfo.g_clients--;
		hash_del_client(cptr);
		_user_del_channels(cptr);
		_channel_invite_del(NULL, cptr);
		if (cptr->servptr) /* has to be true ? */
			del_client_from_ulist(&(cptr->servptr->users), cptr);
		if (cptr->away)
			MyFree(cptr->away);
		MyFree(cptr);
	}
	return 0;
}

/* called for local connections that are unknown or auth */
static int _client_del_local(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason)
{
	assert(IsLocal(cptr));
	if (cptr->connectby)
	{
		MyFree(cptr->connectby);
		cptr->connectby = NULL;
	}
	client_queue_flush(cptr, NULL);
	ircd_close_fd(&(cptr->fdlistinfo));
    queue_delete(&(cptr->queue_in));
    queue_delete(&(cptr->queue_out));
	if (IsRejecthold(cptr))
	{
		del_client_from_lolist(&Rejectholds, cptr);
		MyFree(cptr);
		return 0;
	}
	if (!IsDead(cptr))
	{
		cptr->flags |= CLIENT_FLAGS_DEAD;
#ifndef NODEBUG
		cptr->deadtype = 1;
#endif
	}
	else
		assert(cptr->deadtype != 1);
	assert(IsLocal(cptr));
	del_client_from_lolist(&Locals, cptr);
	if (IsAuth(cptr))
	{
		if (cptr->auth_ptr)
		{
			cptr->auth_ptr->flags &= ~CLIENT_FLAGS_DOINGAUTH;
			cptr->auth_ptr->auth_ptr = NULL;
		}
		MyFree(cptr);
		return 0;
	}
	else
	{
		if (IsServer(cptr))
		{
			int connected = Now - cptr->firsttime;

			send_to_opers("Link with %s closed: %s", cptr->name, reason);
			send_to_opers("%s had been connected for %d day%s, %02d:%02d:%02d",
				cptr->name, connected/86400,
				(connected/86400 == 1) ? "" : "s",
				(connected % 86400) / 3600,
				(connected % 3600) / 60,
				(connected % 60));
		}
		else if (cptr->flags & CLIENT_FLAGS_SCONNECT)
		{
			send_to_opers("Link with %s could not be established: %s", cptr->name, reason);
		}
	}
	DEBUG(DEBUG_ALL, ("_client_del_local(): client disconnecting: %s%s%s@%s\n",
			cptr->name, *cptr->name ? "!" : "",
			cptr->username, cptr->hostname));
	myres_status_deinit(&(cptr->res_status));
	myres_hostent_free(&(cptr->he_forward));
	myres_hostent_free(&(cptr->he_reverse));
    if (cptr->auth_ptr)
    {
		cptr->auth_ptr->auth_ptr = NULL;
		_client_del_local(from, cptr->auth_ptr, NULL);
		cptr->auth_ptr = NULL;
    }
	if (cptr->away)
	{
		MyFree(cptr->away);
		cptr->away = NULL;
	}
	if (cptr->flood_list)
	{
		del_client_from_flist(cptr->flood_list, cptr);
		cptr->flood_list = NULL;
	}
	if (IsReg(cptr))
	{
		whowas_add_history(cptr, 0);
		whowas_off_history(cptr);
		if (cptr->servptr)
		{
			del_client_from_ulist(&(cptr->servptr->users), cptr);
			cptr->servptr = NULL;
		}
		if (cptr->umodes & CLIENT_UMODES_OPER)
		{
			Globalinfo.m_opers--;
			Globalinfo.g_opers--;
		}
		if (cptr->umodes & CLIENT_UMODES_IMODE)
		{
			Globalinfo.m_invis--;
			Globalinfo.g_invis--;
		}
		Globalinfo.m_clients--;
		Globalinfo.g_clients--;
		if (IsClient(cptr))
		{
			int on_for = Now - cptr->firsttime;

			if (vars_value_get_bynum(IRCD_VAR_USERS_LOG))
				log_printf(_ircd_conf.log_users, "%s (%3d:%02d:%02d): %s!%s@%s [%s] [%s]\n",
					support_ctime(cptr->firsttime), on_for/3600,
					(on_for % 3600) / 60, on_for % 60,
					cptr->name, cptr->username, cptr->hostname,
					cptr->authusername, cptr->ipstring);
		}
	}
	else if (IsServer(cptr))
	{
		assert(cptr->servptr == meptr);
		if (cptr->servptr)
		{
			del_client_from_ulist(&(cptr->servptr->servers), cptr);
			cptr->servptr = NULL;
		}
		Globalinfo.m_servers--;
#if 0
		/* done in del_remote */
		Globalinfo.g_servers--;
#endif
	}
	if (cptr->listener)
	{
		del_client_from_llist(&(cptr->listener->listener), cptr);
		if (!--cptr->listener->num_on &&
					(cptr->listener->flags & CLIENT_FLAGS_ILLEGAL))
		{
			assert(cptr->listener->listener == NULL);
			del_client_from_oslist(&(_ircd_conf.listeners), cptr->listener);
			ircd_close_fd(&(cptr->listener->fdlistinfo));
            MyFree(cptr->listener);
		}
		cptr->listener = NULL;
	}
	if (*cptr->name && (IsUnknown(cptr) || IsClient(cptr)))
		hash_del_client(cptr);
	if (cptr->oslist)
	{
		del_client_from_oslist(cptr->oslist, cptr);
		cptr->oslist = NULL;
	}
	if (cptr->ipaddr_ptr)
	{
		iphash_handle_exit(cptr);
	}
	if (cptr->class)
	{
		del_client_from_cllist(&(cptr->class->clients), cptr);
		if (!--cptr->class->num_on &&
					(cptr->class->flags & CLASS_FLAGS_ILLEGAL))
		{
			class_del(&(_ircd_conf.classes), cptr->class);
		}
	}
	if (IsServer(cptr))
	{
		char	comment[HOSTLEN*2+2];

		strcpy(comment, meptr->name);
		strcat(comment, " ");
		strcat(comment, cptr->name);
		_client_del_remote(from, cptr, comment, 1, NULL);
		/* cptr is gone */
		return 0;
	}
	if (IsClient(cptr))
	{
		if (!(cptr->flags & CLIENT_FLAGS_NOSENDQUIT))
		{
			send_to_servers(cptr->from, ":%s QUIT :%s",
				cptr->name, reason);
		}
		if (vars_value_get_bynum(IRCD_VAR_CLIENTCONNECT))
		{
			if (!(cptr->flags & CLIENT_FLAGS_KILLED))
				send_to_flag(CLIENT_NFLAGS_CLIDISCONNECT, NULL,
					"Client exiting: %s (%s@%s) [Client Quit]",
					cptr->name, cptr->username, cptr->hostname);
			else
				send_to_flag(CLIENT_NFLAGS_CLIDISCONNECT, NULL,
					"Client exiting: %s (%s@%s) [%s]",
					cptr->name, cptr->username, cptr->hostname, reason);
		}
		send_to_common_channels(cptr, "QUIT :%s", reason);
		_user_del_channels(cptr);
		_channel_invite_del(NULL, cptr);
	}
    MyFree(cptr);
    return 0;
}

int ircd_parse(IRCD_CLIENT *cptr)
{
	int				len;
	char			buf[512];
	IRCD_CLIENT		*from;
	char			*ptr;
	char			*sender;
	char			*command;
	int				numeric;
	COMMAND			*mptr;
	int				ret;

	if (!cptr)
		return -EINVAL;
	if (!cptr->queue_in.queue_size)
	{
		DEBUG(DEBUG_ALL, ("ircd_parse(): queue_size is 0, returning\n"));
		return 0;
	}
	cptr->flags &= ~CLIENT_FLAGS_PINGSENT;
	for(;;)
	{
#ifdef MALLOC_DEBUG
		{
			static int	lastmalloccheck = 0;

			if (Now - lastmalloccheck > 30)
			{
				dmalloc_check();
				lastmalloccheck = Now;
			}
		}
#endif
		if (cptr->rate_msg_limit && !(IsServer(cptr) ||
			(IsAnOper(cptr) && (cptr->oline_flags & OLINE_FLAGS_CANFLOOD))))
		{
			if (Now > cptr->rate_time)
			{
				cptr->rate_time = Now + cptr->rate_time_limit;
				cptr->rate_msgs = 0;
				if (cptr->flood_list)
				{
					del_client_from_flist(cptr->flood_list, cptr);
					cptr->flood_list = NULL;
				}
			}
			if (cptr->rate_msgs >= cptr->rate_msg_limit)
			{
				if (Now > cptr->rate_lastflood + 300)
					cptr->rate_exit_num = 0;
				cptr->rate_lastflood = Now;
				if (!cptr->flood_list)
				{
					cptr->flood_list = &Flooders;
					add_client_to_flist(cptr->flood_list, cptr);
					cptr->rate_exit_num++;
					if (cptr->rate_exit_num_limit &&
						(cptr->rate_exit_num >= cptr->rate_exit_num_limit))
					{
						return client_del(cptr, cptr, "Too much flooding");
					}
				}
				return 0;
			}
		}

		len = queue_data_get_nl(&(cptr->queue_in), buf, sizeof(buf)-1, 0);
		DEBUG(DEBUG_ALL, ("ircd_parse(): get_nl() returned %i\n", len));
		if (len <= 0)
			return 0;
		DEBUG(DEBUG_ALL, ("%5i: %s\n", len, buf));

		cptr->recv_msg++;

		for(ptr=buf;*ptr==' ';ptr++)
			continue;

		from = NULL;
		sender = NULL;

		if (*ptr == ':')
		{
			sender = ++ptr;
			while(*ptr && (*ptr != ' '))
				ptr++;
			if (!*ptr)
				continue; /* ":<prefix>\0" */
			if (*sender && IsServer(cptr))
			{
				*ptr++ = '\0';
				from = hash_find_client_incl_hm(sender);
				if (!from)
				{
					DEBUG(DEBUG_ALL, ("unknown prefix, \"%s\" from %s\n",
							sender, cptr->name));
					continue;
				}
				if (from->from != cptr)
				{
					DEBUG(DEBUG_ALL, ("prefix from wrong direction, \"%s\" from %s instead of %s\n",
						sender, cptr->name, from->from->name));
					continue;
				}
			}
			else
				sender = cptr->name;
			while(*ptr == ' ')
				ptr++;
		}
		else
			sender = cptr->name;
		if (!*ptr)
			continue;
		command = ptr;
		while(*ptr && (*ptr != ' '))
			ptr++;
		len = ptr - command;
		numeric = 0;
		if ((len == 3) && *ptr &&
				isdigit(*command) &&
				isdigit(*(command+1)) &&
				isdigit(*(command+2)))
		{
			numeric = ((*command - '0')*100)+
						((*(command + 1)-'0')*10)+(*(command + 2)-'0');
			mptr = NULL;
		}
		else
		{
			if (*ptr)
				*ptr++ = '\0';
			while(*ptr && (*ptr == ' '))
				ptr++;
			mptr = _command_tree_find(command);
			if (!mptr)
			{
				if (IsClient(cptr))
					send_prefix_to_client(meptr, cptr,
						"%d %s %s :Unknown command", ERR_UNKNOWNCOMMAND,
						cptr->name, command);
				DEBUG(DEBUG_ALL, ("command not found: %s\n", command));
				cptr->rate_msgs++;
				continue;
			}
			DEBUG(DEBUG_ALL, ("command found: %s\n", command));
		}
		if (!from)
			from = cptr;
		if (mptr)
			ret = mptr->function(cptr, from, sender, ptr);
		else if (numeric)
			ret = do_numeric(cptr, from, numeric, sender, ptr); /* do numeric */
		else
			ret = 0;
		cptr->rate_msgs += ret;
		if (ret < 0)
			return ret;
	}
	return 0;
}

char *ircd_nextarg(char **src, int flags)
{
	char	*tok;

	if (!src || !*src || !**src)
		return NULL;
	while(**src && **src == ' ')
		(*src)++;
	if (**src)
		tok = *src;
	else
		return NULL;
	if (*tok == ':')
	{
		tok++;
		*src = NULL;
	}
	else if (!(flags & NEXTARG_FLAGS_LASTARG))
	{
		while(**src && **src != ' ')
			(*src)++;
		if (**src)
		{
			**src = '\0';
			(*src)++;
			while(**src && **src == ' ')
				(*src)++;
		}
	}
	else
		*src = NULL;
	return tok;
}

char *ircd_clientname(IRCD_CLIENT *sptr, int showip, char *buffer)
{
	if (IsLocal(sptr))
	{
		if (showip)
			sprintf(buffer, "%s[%s@%s]",
				sptr->name, *sptr->authusername ? sptr->authusername : "",
				sptr->ipstring);
		else
		{
			if (IsServer(sptr))
				strcpy(buffer, sptr->name);
			else if (*sptr->hostname)
				sprintf(buffer, "%s[%s%s%s]",
					sptr->name, *sptr->username ? sptr->username : "",
					*sptr->username ? "@" : "", sptr->hostname);
			else
				sprintf(buffer, "%s[%s%s%s]",
					sptr->name, *sptr->username ? sptr->username : "",
					*sptr->username ? "@" : "",
					*sptr->servername ? sptr->servername : "UNKNOWN");
		}
	}
	else
		strcpy(buffer, sptr->name);
	return buffer;
}

int client_queue_flush(IRCD_CLIENT *cptr, char *exitbuf)
{
	int ret;

	for(;cptr->queue_out.queue_size;)
	{
		char	*ptr;
		int		len;

		ret = queue_data_get(&(cptr->queue_out), &ptr, &len);
		assert(ret >= 0);
		ret = write(cptr->fdlistinfo.fd, ptr, len);
		if (ret > 0)
		{
			if (IsClient(cptr))
				Globalinfo.bw_c_outgoing_bytes += ret;
			else if (IsServer(cptr))
				Globalinfo.bw_s_outgoing_bytes += ret;
			cptr->writetries = 0;
			queue_data_del(&(cptr->queue_out), ret);
			if (ret == len)
				continue;
		}
		else if (!ret)
		{
			cptr->flags |= CLIENT_FLAGS_DEAD;
			cptr->deadtype = 2;
			if (exitbuf)
				strcpy(exitbuf, "Write error: EOF of client");
			return -1;
		}
		else if (ret < 0)
		{
			if ((errno != EAGAIN) && (errno != EINTR) &&
					(errno != EWOULDBLOCK))
			{
				if (exitbuf)
					sprintf(exitbuf, "Write error: %i", errno);
				return -1;
			}
		}
		break;
	}
	if (!(cptr->lastqsize = cptr->queue_out.queue_size))
		fdlist_del_event(cptr->fdlistinfo.fdlist,
			cptr->fdlistinfo.fdlist_pos, FDLIST_FLAGS_WRITE);
	return 0;
}

int ircd_ratelimit(IRCD_CLIENT *pptr, int *numptr, int *timeptr, int numlimit, int timelimit)
{
	if (!numlimit || !timelimit || !IsClient(pptr))
		return 0;
	if (Now - *timeptr > timelimit)
	{
		*numptr = 1;
		*timeptr = Now;
		return 0;
	}
	if (++(*numptr) <= numlimit)
		return 0;
	if (*numptr == (numlimit + 1))
		*timeptr = Now; /* wait 'timelimit' seconds before allowing another*/
	return 1;
}

int debug_locals(void)
{
	IRCD_CLIENT *lptr;
	char buffer[1024];
	int fd;

	fd = open("locals.dbg", O_CREAT|O_WRONLY|O_APPEND, 0660);
	if (fd >= 0)
	{
		for(lptr=Locals;lptr;lptr=lptr->lo_next)
		{
			sprintf(buffer, "%x: %s\n", (int)lptr, lptr->name);
			write(fd, buffer, strlen(buffer));
			fsync(fd);
		}
		close(fd);
	}
	return 0;
}
